// -*- C++ -*-

#pragma once


#include "range.h"
#include <exception>
#include <iostream>
#include <map>
#include <memory>


inline void goopax_debugfunc()
{
}


namespace goopax {
namespace impl {


#if GOOPAX_IOS && GOOPAX_IOS_VERSION < 90000
template<typename = void>
struct exception_holder
{
    static std::map<thread::id, std::unique_ptr<std::exception_ptr>> current_exceptions;
    static inline std::exception_ptr&                                current_exception()
    {
        static mutex      Exmutex;
        lock_guard<mutex> LOCK(Exmutex);
        auto&             e = current_exceptions[this_thread::get_id()];
        if (e.get() == nullptr)
        {
            e = std::make_unique<std::exception_ptr>();
        }
        return *e;
    }
};
template<typename T>
std::map<thread::id, std::unique_ptr<std::exception_ptr>> exception_holder<T>::current_exceptions;
#else
template<typename = void>
struct exception_holder
{
    thread_local static std::exception_ptr current_exception_i;
    static inline std::exception_ptr&      current_exception()
    {
        return current_exception_i;
    }
};
template<typename T>
thread_local std::exception_ptr exception_holder<T>::current_exception_i;
#endif


struct EX
{
    /// base exception thrown by goopax
    struct goopax_exception : public std::exception
    {
        public:
        virtual const char* what() const noexcept final
        {
            return interface::goopax_4_ex_what();
        }
        goopax_exception() = default;
    };

    struct general : public goopax_exception
    {
    };


    struct integrity : public goopax_exception
    {
    };

    struct syntax : public goopax_exception
    {
    };

    struct memory : public goopax_exception
    {
    };

    struct assertion : public goopax_exception
    {
    };

    struct backend : public goopax_exception
    {
    };

    struct user : public goopax_exception
    {
    };


    template<class = void>
    static void throw_exception(errno_type err)
    {
#ifdef _WIN32
        std::cerr << "GOOPAX: Got exception '" << interface::goopax_4_ex_what() << "'" << std::endl;
#endif

#define MAKEEX(ex, num) \
    if (err == num) { throw EX::ex(); }

        MAKEEX(general, ERR_general)
        MAKEEX(integrity, ERR_integrity)
        MAKEEX(syntax, ERR_syntax)
        MAKEEX(memory, ERR_memory)
        MAKEEX(assertion, ERR_assertion)
        MAKEEX(backend, ERR_backend)

#undef MAKEEX

        if (err == ERR_user)
        {
            std::exception_ptr ex;
            std::swap(ex, exception_holder<>::current_exception());
            std::rethrow_exception(std::move(ex));
        }

        std::cerr << "Unknown errno=" << err << "!" << std::endl;
        abort();
    }
};

inline void check_errno(errno_type err)
{
    if (GOOPAX_EXPECT(err, SUCCESS) != SUCCESS)
    {
        EX::throw_exception(err);
    }
}

template<class R, class... A, class... B>
R if_call(errno_type (*func)(R*, A...) noexcept, B... a)
{
    R r;
    check_errno((*func)(&r, a...));
    return r;
}
}
using impl::EX;
}
